from typing import Any, Dict, List, Literal, NamedTuple, Optional, Union, cast

from django import forms
from django.core.exceptions import ValidationError
from django.forms.models import ModelChoiceIterator

from .serialization.registry import register
from .serialization.widgets import BaseWidget


class Autocomplete(forms.Select):
    template_name: Literal["reactivated/autocomplete"] = "reactivated/autocomplete"

    def get_context(
        self, name: str, value: Any, attrs: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        choices = cast(ModelChoiceIterator, self.choices)

        assert choices.queryset is not None
        assert hasattr(
            choices.queryset, "autocomplete"
        ), "Models marked for autocompletion must implement autocomplete(query: str) at the manager level"

        to_field_name = choices.field.to_field_name or "pk"

        # context = forms.Widget.get_context(self, name, value, attrs)
        # self.choices.queryset = self.choices.queryset._clone()[:10]
        # context = super().get_context(name, value, attrs)
        context = forms.Widget.get_context(self, name, value, attrs)
        try:
            selected = choices.field.to_python(value)
        except ValidationError:
            selected = None

        if selected is not None:
            context["widget"]["selected"] = {
                "value": getattr(selected, to_field_name),
                "label": str(selected),
            }
        else:
            context["widget"]["selected"] = None

        return context


class AutocompleteSelected(NamedTuple):
    value: Union[str, int]
    label: str


@register(Autocomplete)
class AutocompleteType(BaseWidget):
    value: List[str]  # type: ignore[assignment]
    selected: Optional[AutocompleteSelected]



import importlib
import json
import logging
import os
import subprocess
from typing import Any, Dict, NamedTuple, Tuple

from django.apps import AppConfig
from django.conf import settings
from django.utils.autoreload import file_changed  # type: ignore[attr-defined]

from . import extract_views_from_urlpatterns, types
from .serialization import create_schema
from .serialization.registry import (
    definitions_registry,
    global_types,
    template_registry,
    type_registry,
    value_registry,
)

logger = logging.getLogger("django.server")


def get_urls_schema() -> Dict[str, Any]:
    urlconf = importlib.import_module(settings.ROOT_URLCONF)
    urlpatterns = urlconf.urlpatterns  # type: ignore[attr-defined]

    from django.urls import converters
    from django.urls.resolvers import RoutePattern

    converter_mapping = {
        converters.IntConverter: "number",
        converters.StringConverter: "string",
        converters.UUIDConverter: "string",
        converters.SlugConverter: "string",
        converters.PathConverter: "string",
    }

    urls = extract_views_from_urlpatterns(urlpatterns)  # type: ignore[no-untyped-call]
    reverse: types.URLSchema = {}

    for _, regex, name, pattern in urls:
        if not isinstance(pattern, RoutePattern):
            continue
        reverse[name or regex] = {
            "route": f"/{regex}",
            "args": {
                arg_name: converter_mapping.get(arg_converter.__class__, "string")
                for arg_name, arg_converter in pattern.converters.items()
            },
        }

    return reverse


def get_types_schema() -> Any:
    """The package json-schema-to-typescript does expose a way to
    automatically export any interface it sees. However, this can bloat our
    generated files.

    Instead, while creating the schema, we occasionally run into types that we
    want available globally but are not directly referenced by templates.

    These aren't exported by `json-schem-to-typescript` because they're
    referenced using `tsType`, so the libraary is unaware of their usage.

    So we register them in `globals` and force `json-schema-to-typescript` to
    expose them.

    We can't just add these types to the `type_registry` because that's only
    parsed once when generating the parent tuple.

    We could explore doing two passes in the future.

    See `unreachableDefinitions` in json-schema-to-typescript
    """
    type_registry["globals"] = Any  # type: ignore[assignment]

    context_processors = []

    from .serialization.context_processors import create_context_processor_type

    for engine in settings.TEMPLATES:
        if engine["BACKEND"] == "reactivated.backend.JSX":
            context_processors.extend(engine["OPTIONS"]["context_processors"])  # type: ignore[index]

    type_registry["Context"] = create_context_processor_type(context_processors)

    ParentTuple = NamedTuple("ParentTuple", type_registry.items())  # type: ignore[misc]
    parent_schema, definitions = create_schema(ParentTuple, definitions_registry)
    definitions_registry.update(definitions)

    return {
        "definitions": definitions,
        **{
            **definitions["reactivated.apps.ParentTuple"],
            "properties": {
                **definitions["reactivated.apps.ParentTuple"]["properties"],
                "globals": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": list(global_types.keys()),
                    "properties": global_types,
                },
            },
        },
    }


def get_templates() -> Dict[str, Tuple[Any]]:
    return template_registry


def get_values() -> Dict[str, Any]:
    return value_registry


def get_schema() -> str:
    schema = {
        "urls": get_urls_schema(),
        "templates": get_templates(),
        "types": get_types_schema(),
        "values": get_values(),
    }
    return json.dumps(schema, indent=4)


class ReactivatedConfig(AppConfig):
    name = "reactivated"

    def ready(self) -> None:
        """
        Django's dev server actually starts twice. So we prevent generation on
        the first start. TODO: handle noreload.
        """

        from .checks import check_installed_app_order  # NOQA
        from .serialization import widgets  # noqa

        def regenerate_schema(*args: Any, **kwargs: Any) -> None:
            schema = get_schema()
            generate_schema(schema)

        file_changed.connect(regenerate_schema)


def generate_schema(schema: str, skip_cache: bool = False) -> None:
    """
    For development usage only, this requires Node and Python installed

    You can use this function for your E2E test prep.
    """
    logger.info("Generating interfaces and client side code")
    encoded_schema = schema.encode()

    import hashlib

    digest = hashlib.sha1(encoded_schema).hexdigest().encode()

    if skip_cache is False and os.path.exists(
        f"{settings.BASE_DIR}/client/generated/index.tsx"
    ):
        with open(f"{settings.BASE_DIR}/client/generated/index.tsx", "r+b") as existing:
            already_generated = existing.read()

            if digest in already_generated:
                logger.info("Skipping generation as nothing has changed")
                return

    #: Note that we don't pass the file object to stdout, because otherwise
    # webpack gets confused with the half-written file when we make updates.
    # Maybe there's a way to force it to be a single atomic write? I tried
    # open('w+b', buffering=0) but no luck.
    process = subprocess.Popen(
        ["node", f"{settings.BASE_DIR}/node_modules/reactivated/generator.js"],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        cwd=settings.BASE_DIR,
    )
    out, error = process.communicate(encoded_schema)

    constants_process = subprocess.Popen(
        [
            "node",
            f"{settings.BASE_DIR}/node_modules/reactivated/generator/constants.js",
        ],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        cwd=settings.BASE_DIR,
    )
    constants_out, constants_error = constants_process.communicate(encoded_schema)

    os.makedirs(f"{settings.BASE_DIR}/client/generated", exist_ok=True)

    with open(f"{settings.BASE_DIR}/client/generated/index.tsx", "w+b") as output:
        output.write(b"// Digest: %s\n" % digest)
        output.write(out)

    with open(f"{settings.BASE_DIR}/client/generated/constants.tsx", "w+b") as output:
        output.write(constants_out)

    logger.info("Finished generating.")



from typing import Any, Dict, NamedTuple, Optional, Type, TypeVar

from django import forms
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.template.response import TemplateResponse

from . import utils
from .renderer import should_respond_with_json
from .serialization import create_schema, serialize
from .serialization.registry import (
    JSON,
    definitions_registry,
    template_registry,
    type_registry,
)

T = TypeVar("T", bound=NamedTuple)


class LazySerializationResponse(TemplateResponse):
    """
    This lazy serialization doesn't actually convert our context to JSON
    until the very last minute, using resolve_context.

    That allows things like our autocomplete decorator and other functions
    that peek into the context to behave as normal. They'll be handed
    regular Python objects.
    """

    _request: HttpRequest

    def __init__(
        self,
        request: HttpRequest,
        template: Type[T],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.template = template
        super().__init__(request, *args, **kwargs)

    @property
    def rendered_content(self) -> str:
        from .backend import JSXTemplate

        engine = utils.get_template_engine()
        template = JSXTemplate(self.template_name, engine)  # type: ignore[arg-type]
        context = self.resolve_context(self.context_data)
        return template.render(context, self._request)

    def __getstate__(self) -> Any:
        """
        First run the normal pickling of `TemplateResponse`. That already has
        special handling.

        Then, remove our template from pickling because it cannot and should
        not be pickled. The response only cares about the rendered content.
        """
        obj_dict = super().__getstate__()  # type: ignore[misc]
        del obj_dict["template"]
        return obj_dict

    def resolve_context(self, context: Optional[Dict[str, Any]]) -> JSON:

        generated_schema = create_schema(self.template, definitions_registry)
        return serialize(context, generated_schema)


def template(cls: Type[T]) -> Type[T]:
    type_name = f"{cls.__name__}Props"

    class Augmented(cls):  # type: ignore[misc, valid-type]
        @staticmethod
        def register() -> None:
            type_registry[type_name] = cls  # type: ignore[assignment]
            template_registry[cls.__name__] = type_name  # type: ignore[assignment]

        def items(self) -> Any:
            """
            Duck-typing for context, so you can loop over a template.

            See autocomplete.
            """
            return self._asdict().items()

        def get_serialized(self) -> Any:
            generated_schema = create_schema(cls, {})
            return serialize(self, generated_schema)

        def render(self, request: HttpRequest) -> TemplateResponse:
            response = LazySerializationResponse(  # type
                request,
                cls,
                f"{self.__class__.__name__}.tsx",
                self,
            )
            return response

    Augmented.register()
    Augmented.__name__ = cls.__name__
    return Augmented


class Action(NamedTuple):
    name: str


class Extracted(NamedTuple):
    context_forms: Dict[str, forms.BaseForm]
    context_form_sets: Dict[str, forms.BaseFormSet]
    context_actions: Dict[str, Action]


def extract_forms_form_sets_and_actions(interface: Any) -> Extracted:
    context = interface._asdict()

    context_forms = {}
    context_form_sets = {}
    context_actions = {}

    for name, item in context.items():
        if isinstance(item, forms.BaseForm):
            context_forms[name] = item

        elif isinstance(item, forms.BaseFormSet):
            context_form_sets[name] = item

        elif isinstance(item, Action):
            context_actions[name] = item
        elif getattr(item, "is_reactivated_interface", None) is True:
            children = extract_forms_form_sets_and_actions(item)
            context_forms.update(children.context_forms)
            context_form_sets.update(children.context_form_sets)
            context_actions.update(children.context_actions)

    return Extracted(
        context_forms=context_forms,
        context_form_sets=context_form_sets,
        context_actions=context_actions,
    )


def interface(cls: T) -> T:
    type_name = f"{cls.__name__}Props"  # type: ignore[attr-defined]
    type_registry[type_name] = cls  # type: ignore[assignment]

    class Augmented(cls):  # type: ignore[misc, valid-type]
        is_reactivated_interface = True

        def get_serialized(self) -> Any:
            generated_schema = create_schema(cls, definitions_registry)
            return serialize(self, generated_schema)

        def render(self, request: HttpRequest) -> HttpResponse:
            import simplejson

            serialized = simplejson.dumps(self.get_serialized(), indent=4)

            if should_respond_with_json(request):
                return HttpResponse(serialized, content_type="application/json")

            extracted_context = extract_forms_form_sets_and_actions(self)

            return TemplateResponse(
                request,
                "reactivated/interface.html",
                {
                    **self._asdict(),
                    "forms": extracted_context.context_forms,
                    "form_sets": extracted_context.context_form_sets,
                    "actions": extracted_context.context_actions,
                    "serialized": serialized,
                },
            )

        def as_json(self, request: HttpRequest) -> JsonResponse:
            return JsonResponse(self.get_serialized())

    Augmented.__name__ = cls.__name__  # type: ignore[attr-defined]
    return Augmented  # type: ignore[return-value]
